﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.SimulationCore {
  [Item("SortedListEventQueue", "")]
  [StorableClass]
  public sealed class SortedListEventQueue<TModel> : Item, IEventQueue<TModel> where TModel : class, IModel {
    [Storable]
    private readonly SortedList<double, List<IEvent<TModel>>> queue = new SortedList<double, List<IEvent<TModel>>>();

    public bool Empty { get { return queue.Count == 0; } }
    public int Count { get { return queue.Count; } }

    public IEvent<TModel> Peek() {
      return queue.First().Value.First();
    }

    public IEvent<TModel> Pop() {
      var @event = queue.First().Value.First();
      Remove(@event);
      return @event;
    }

    public void Push(double time, IAction<TModel> action) {
      PushEvent(time, action);
    }

    private IEvent<TModel> PushEvent(double time, IAction<TModel> action) {
      IEvent<TModel> result = new Event<TModel>(time, action);
      if (queue.ContainsKey(time))
        queue[time].Add(result);
      else queue.Add(time, new List<IEvent<TModel>>() { result });
      return result;
    }

    public IEvent<TModel> ChangeTime(IEvent<TModel> @event, double newTime) {
      Remove(@event);
      return PushEvent(newTime, @event.Action);
    }

    public void Delete(IEvent<TModel> @event) {
      Remove(@event);
    }

    private void Remove(IEvent<TModel> @event) {
      if (!queue[@event.Time].Remove(@event)) throw new InvalidOperationException("Event is not in queue.");
      if (queue[@event.Time].Count == 0) queue.Remove(@event.Time);
    }

    public void Clear() {
      queue.Clear();
    }

    [StorableConstructor]
    private SortedListEventQueue(bool deserializing) : base(deserializing) { }
    private SortedListEventQueue(SortedListEventQueue<TModel> original, Cloner cloner)
      : base(original, cloner) {
      queue = new SortedList<double, List<IEvent<TModel>>>(
        original.queue
        .Select(x => new KeyValuePair<double, List<IEvent<TModel>>>(x.Key, x.Value.Select(cloner.Clone).ToList()))
        .ToDictionary(x => x.Key, x => x.Value));
    }
    public SortedListEventQueue() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SortedListEventQueue<TModel>(this, cloner);
    }
  }
}
